@@ -1,10 +1,9 @@
-/* Copyright (C) 1997, 1998, 1999, 2000, 2002, 2003 Thorsten Kukuk
+/* Copyright (C) 1997-2016 Thorsten Kukuk
    Author: Thorsten Kukuk <kukuk@suse.de>
 
    The YP Server is free software; you can redistribute it and/or
-   modify it under the terms of the GNU General Public License as
-   published by the Free Software Foundation; either version 2 of the
-   License, or (at your option) any later version.
+   modify it under the terms of the GNU General Public License
+   version 2 as published by the Free Software Foundation.
 
    The YP Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -13,33 +12,34 @@
 
    You should have received a copy of the GNU General Public
    License along with the YP Server; see the file COPYING. If
-   not, write to the Free Software Foundation, Inc., 675 Mass Ave,
-   Cambridge, MA 02139, USA. */
+   not, write to the Free Software Foundation, Inc., 51 Franklin Street,
+   Suite 500, Boston, MA 02110-1335, USA. */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
 #include <netdb.h>
+#include <unistd.h>
 #include <syslog.h>
-#ifndef LOG_DAEMON
-#include <sys/syslog.h>
-#endif
 #include <string.h>
 #include <stdlib.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#if USE_SD_NOTIFY
+#include <systemd/sd-daemon.h>
+#endif
 
 #include "log_msg.h"
 #include "ypserv_conf.h"
 #include "access.h"
 #include "yp_db.h"
 #include "yp.h"
-#include "compat.h"
 
 static conffile_t *conf = NULL;
+const char *confdir = CONFDIR;
 
 void
 load_config (void)
@@ -48,7 +48,7 @@ load_config (void)
 
   if (conf != NULL)
     {
-      log_msg ("Reloading %s/ypserv.conf", CONFDIR);
+      log_msg ("Reloading %s/ypserv.conf", confdir);
       while (conf)
 	{
 	  tmp = conf;
@@ -59,7 +59,7 @@ load_config (void)
 	}
     }
 
-  conf = load_ypserv_conf (CONFDIR);
+  conf = load_ypserv_conf (confdir);
 }
 
 /* Give a string with the DEFINE description back */
@@ -92,7 +92,10 @@ ypproc_name (int proc)
       return "ypproc_order";
     case YPPROC_MAPLIST:
       return "ypproc_maplist";
+    case YPPROC_NEWXFR:
+      return "ypproc_newxfr";
     default:
+      log_msg ("Unknown procedure '%i'", proc);
       return "unknown ?";
     }
 }
@@ -119,6 +122,50 @@ is_valid_domain (const char *domain)
   return 1;
 }
 
+static struct netbuf *
+copy_netbuf (struct netbuf *src)
+{
+  struct netbuf *dst;
+
+  if (src == NULL)
+    return NULL;
+
+  dst = calloc (1, sizeof (struct netbuf));
+  if (dst == NULL)
+    return NULL;
+
+  dst->buf = malloc (src->len);
+  if (dst->buf == NULL)
+    {
+      free (dst);
+      return NULL;
+    }
+
+  dst->len = src->len;
+  dst->maxlen = src->len;
+  memcpy (dst->buf, src->buf, src->len);
+
+  return dst;
+}
+
+/* return values:
+   0: both IP addresses are identical
+   1: the IP addresses are not identical */
+static int
+cmp_netbuf (struct netbuf *nbuf1, struct netbuf *nbuf2)
+{
+  if (nbuf1 == NULL || nbuf2 == NULL)
+    return 1;
+
+  if (nbuf1->len != nbuf2->len)
+    return 1;
+
+  if (memcmp (nbuf1->buf, nbuf2->buf, nbuf1->len) == 0)
+    return 0;
+
+  return 1;
+}
+
 /* By default, we use the securenet list, to check if the client
    is secure.
 
@@ -126,15 +173,17 @@ is_valid_domain (const char *domain)
    return  0, if securenets does not allow access from this host
    return -1, if request comes from an unauthorized host
    return -2, if the map name is not valid
-   return -3, if the domain is not valid */
-
+   return -3, if the domain is not valid
+   return -4, if the map does not exist */
 int
 is_valid (struct svc_req *rqstp, const char *map, const char *domain)
 {
-  const struct sockaddr_in *sin;
-  int status;
-  static unsigned long int oldaddr = 0;		/* so we dont log multiple times */
+  static struct netbuf *oldaddr = NULL;  /* so we dont log multiple times */
   static int oldstatus = -1;
+  struct netconfig *nconf;
+  struct netbuf *rqhost;
+  struct __rpc_sockinfo si;
+  int status;
 
   if (domain && is_valid_domain (domain) == 0)
     return -3;
@@ -142,12 +191,17 @@ is_valid (struct svc_req *rqstp, const char *map, const char *domain)
   if (map && (map[0] == '\0' || strchr (map ,'/')))
     return -2;
 
-  sin = svc_getcaller (rqstp->rq_xprt);
+  rqhost = svc_getrpccaller (rqstp->rq_xprt);
+  nconf = getnetconfigent (rqstp->rq_xprt->xp_netid);
 
-  status = securenet_host (sin->sin_addr);
+  status = securenet_host (nconf, rqhost);
 
-  if ((map != NULL) && status)
+  if (!__rpc_nconf2sockinfo(nconf, &si))
+    return -1;
+
+  if ((map != NULL) && status && si.si_af == AF_INET)
     {
+      struct sockaddr_in *sin = rqhost->buf;
       conffile_t *work;
 
       work = conf;
@@ -170,7 +224,7 @@ is_valid (struct svc_req *rqstp, const char *map, const char *domain)
 	    status = -1;
 	    break;
 	  case SEC_PORT:
-	    if (ntohs (sin->sin_port) >= IPPORT_RESERVED)
+	    if (taddr2port (nconf, rqhost) >= IPPORT_RESERVED)
 	      status = -1;
 	    break;
 	  }
@@ -186,30 +240,73 @@ is_valid (struct svc_req *rqstp, const char *map, const char *domain)
 	      key.dsize = sizeof ("YP_SECURE") - 1;
 	      key.dptr = "YP_SECURE";
 	      if (ypdb_exists (dbp, key))
-		if (ntohs (sin->sin_port) >= IPPORT_RESERVED)
+		if (taddr2port (nconf, rqhost) >= IPPORT_RESERVED)
 		  status = -1;
 	      ypdb_close (dbp);
 	    }
+          else
+              status = -4;
 	}
     }
 
   if (debug_flag)
     {
-      log_msg ("%sconnect from %s", status ? "" : "refused ",
-	       inet_ntoa (sin->sin_addr));
+      char host[INET6_ADDRSTRLEN];
+
+      log_msg ("%sconnect from %s port %d to procedure %s (%s,%s;%d)",
+	       status ? "" : "refused ",
+	       taddr2ipstr (nconf, rqhost, host, sizeof (host)),
+	       taddr2port (nconf, rqhost), ypproc_name (rqstp->rq_proc),
+	       domain ? domain : "", map ? map : "", status);
     }
   else
     {
-      if (status < 1 && ((sin->sin_addr.s_addr != oldaddr)
-			 || (status != oldstatus)))
-	syslog (LOG_WARNING,
-		"refused connect from %s:%d to procedure %s (%s,%s;%d)\n",
-		inet_ntoa (sin->sin_addr), ntohs (sin->sin_port),
-		ypproc_name (rqstp->rq_proc),
-		domain ? domain : "", map ? map : "", status);
+      if ((status < 1 && status != -4) &&
+	  (cmp_netbuf (oldaddr, rqhost) || (status != oldstatus)))
+	{
+	  char host[INET6_ADDRSTRLEN];
+
+	  syslog (LOG_WARNING,
+		  "refused connect from %s port %d to procedure %s (%s,%s;%d)",
+		  taddr2ipstr (nconf, rqhost, host, sizeof (host)),
+		  taddr2port (nconf, rqhost), ypproc_name (rqstp->rq_proc),
+		  domain ? domain : "", map ? map : "", status);
+	}
     }
-  oldaddr = sin->sin_addr.s_addr;
+
+  /* Create a copy of the netbuf and free the old one */
+  if (oldaddr)
+    {
+      if (oldaddr->buf)
+	free (oldaddr->buf);
+      free (oldaddr);
+    }
+  oldaddr = copy_netbuf (rqhost);
   oldstatus = status;
 
+  freenetconfigent (nconf);
+
   return status;
 }
+
+/* Send a messages to systemd daemon, that inicialization of daemon
+   is finished and daemon is ready to accept connections.
+   It is a nop if we don't use systemd. */
+void
+announce_ready()
+{
+#if USE_SD_NOTIFY
+  int result;
+
+  result = sd_notifyf(0, "READY=1\n"
+                         "STATUS=Processing requests...\n"
+                         "MAINPID=%lu", (unsigned long) getpid());
+
+  /* Return code from sd_notifyf can be ignored, as per sd_notifyf(3).
+     However, if we use systemd's native unit file, we need to send
+     this message to let systemd know that daemon is ready.
+     Thus, we want to know that the call had some issues. */
+  if (result < 0)
+    log_msg ("sd_notifyf failed: %s\n", strerror(-result));
+#endif
+}
