@@ -1,4 +1,4 @@
-/* Copyright (c) 1996, 1997, 1998, 1999, 2000, 2001, 2003 Thorsten Kukuk
+/* Copyright (c) 1996, 1997, 1998, 1999, 2000, 2001, 2003, 2004, 2006, 2008, 2011, 2013, 2014 Thorsten Kukuk
    Author: Thorsten Kukuk <kukuk@suse.de>
 
    The YP Server is free software; you can redistribute it and/or
@@ -12,15 +12,13 @@
 
    You should have received a copy of the GNU General Public
    License along with the YP Server; see the file COPYING. If
-   not, write to the Free Software Foundation, Inc., 675 Mass Ave,
-   Cambridge, MA 02139, USA. */
+   not, write to the Free Software Foundation, Inc., 51 Franklin Street,
+   Suite 500, Boston, MA 02110-1335, USA. */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#define _GNU_SOURCE
-
 #include <stdio.h>
 #include <ctype.h>
 #include <stdlib.h>
@@ -28,19 +26,15 @@
 #include <sys/types.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
-#ifdef HAVE_ALLOCA_H
 #include <alloca.h>
-#endif /* HAVE_ALLOCA_H */
 #include <unistd.h>
 
 #include "log_msg.h"
 #include "ypserv_conf.h"
-#include "compat.h"
 
 int dns_flag = 0;
-#if USE_SLP
 int slp_flag = 0;
-#endif
+unsigned long int slp_timeout = 3600;
 int xfr_check_port = 0;
 char *trusted_master = NULL;
 /* cached_filehandles (how many databases will be cached):
@@ -65,7 +59,7 @@ getipnr (char *n, char *network, char *netmask)
 
   m = strtok (n, "/");
 
-  sscanf (m, "%s", buf);
+  sscanf (m, "%19s", buf);
 
   for (i = 0; i < strlen (buf); i++)
     if ((buf[i] < '0' || buf[i] > '9') && buf[i] != '.')
@@ -172,8 +166,9 @@ conffile_t *
 load_ypserv_conf (const char *path)
 {
   FILE *in;
-  char c, *cp;
-  char buf1[1025], buf2[1025], buf3[1025];
+  int c;
+  char *cp;
+  char buf1[1025], buf2[1025] = "", buf3[1025];
   long line = 0;
   conffile_t *ptr = NULL, *work = NULL;
   char *filename = alloca (strlen (path) + sizeof ("/ypserv.conf") + 1);
@@ -187,7 +182,7 @@ load_ypserv_conf (const char *path)
       return NULL;
     }
 
-  while ((c = fgetc (in)) != (char) EOF)
+  while ((c = fgetc (in)) != EOF)
     {				/*while */
       line++;
       switch (tolower (c))
@@ -198,7 +193,12 @@ load_ypserv_conf (const char *path)
 	    size_t i, j;
 	    unsigned long files = 30;
 
-	    fgets (buf1, sizeof (buf1) - 1, in);
+	    if (fgets (buf1, sizeof (buf1) - 1, in) == NULL)
+	      {
+		log_msg ("Read error in line %d => Ignore line", line);
+		break;
+	      }
+
 	    i = 0;
 	    buf1[sizeof (buf1) - 1] = '\0';
 	    while (c != ':' && i <= strlen (buf1))
@@ -243,7 +243,12 @@ load_ypserv_conf (const char *path)
 	  {
 	    size_t i, j;
 
-	    fgets (buf1, sizeof (buf1) - 1, in);
+	    if (fgets (buf1, sizeof (buf1) - 1, in) == NULL)
+	      {
+		log_msg ("Read error in line %d => Ignore line", line);
+		break;
+	      }
+
 	    i = 0;
 	    while (c != ':' && i <= strlen (buf1))
 	      {
@@ -292,7 +297,12 @@ load_ypserv_conf (const char *path)
 	  {			/* sunos_kludge / slp */
 	    size_t i;
 
-	    fgets (buf1, sizeof (buf1) - 1, in);
+	    if (fgets (buf1, sizeof (buf1) - 1, in) == NULL)
+	      {
+		log_msg ("Read error in line %d => Ignore line", line);
+		break;
+	      }
+
 	    i = 0;
 	    while (c != ':' && i <= strlen (buf1))
 	      {
@@ -315,7 +325,6 @@ load_ypserv_conf (const char *path)
 
 	    if ((buf1[i - 1] == ':') && (strcasecmp (buf2, "slp") == 0))
 	      {
-#if USE_SLP
 		size_t j;
 
 		while (((buf1[i] == ' ') || (buf1[i] == '\t')) &&
@@ -334,20 +343,38 @@ load_ypserv_conf (const char *path)
 		else if (strcasecmp (buf2, "no") == 0)
 		  slp_flag = 0;
 		else
-		  log_msg ("Unknown dns option in line %d: => Ignore line",
+		  log_msg ("Unknown slp option in line %d: => Ignore line",
+			   line);
+
+		if (debug_flag)
+		  log_msg ("ypserv.conf: slp: %d", slp_flag);
+#if !USE_SLP
+		if (slp_flag != 0)
+		  log_msg ("Support for SLP (line %d) is not compiled in.",
 			   line);
-#else
-		log_msg ("Support for SLP (line %d) is not compiled in.",
-			 line);
 #endif
 	      }
+	    else if ((buf1[i - 1] == ':') &&
+		     (strcasecmp (buf2, "slp_timeout") == 0))
+	      {
+		size_t j;
+
+		while (((buf1[i] == ' ') || (buf1[i] == '\t')) &&
+		       (i <= strlen (buf1)))
+		  i++;
+		j = 0;
+		while ((buf1[i] != '\0') && (buf1[i] != '\n'))
+		  buf3[j++] = buf1[i++];
+		buf3[j] = 0;
+
+		sscanf (buf3, "%lu", &slp_timeout);
+
+		if (debug_flag)
+		  log_msg ("ypserv.conf: slp_timeout: %lu", slp_timeout);
+	      }
 	    else
 	      log_msg ("Parse error in line %d: => Ignore line", line);
 
-#if USE_SLP
-	    if (debug_flag)
-	      log_msg ("ypserv.conf: slp: %d", slp_flag);
-#endif
 	    break;
 	  }
 	case 'T':
@@ -355,7 +382,12 @@ load_ypserv_conf (const char *path)
 	  {			/* tryresolve / trusted_master */
 	    size_t i, j;
 
-	    fgets (buf1, sizeof (buf1) - 1, in);
+	    if (fgets (buf1, sizeof (buf1) - 1, in) == NULL)
+	      {
+		log_msg ("Read error in line %d => Ignore line", line);
+		break;
+	      }
+
 	    i = 0;
 	    while (c != ':' && i <= strlen (buf1))
 	      {
@@ -402,7 +434,12 @@ load_ypserv_conf (const char *path)
 	  {			/* xfr_check_port */
 	    size_t i, j;
 
-	    fgets (buf1, sizeof (buf1) - 1, in);
+	    if (fgets (buf1, sizeof (buf1) - 1, in) == NULL)
+	      {
+		log_msg ("Read error in line %d => Ignore line", line);
+		break;
+	      }
+
 	    i = 0;
 	    while (c != ':' && i <= strlen (buf1))
 	      {
@@ -452,7 +489,11 @@ load_ypserv_conf (const char *path)
 	    conffile_t *tmp;
 
 	    buf1[0] = c;
-	    fgets (&buf1[1], sizeof (buf1) - 2, in);
+	    if (fgets (&buf1[1], sizeof (buf1) - 2, in) == NULL)
+	      {
+		log_msg ("Read error in line %d => Ignore line", line);
+		break;
+	      }
 
 	    n = strtok (buf1, ":");
 	    if (n == NULL)
@@ -574,10 +615,11 @@ load_ypserv_conf (const char *path)
 	case '\n':
 	  break;		/* Ignore newline */
 	case '#':
-	  fgets (buf1, sizeof (buf1) - 1, in);
+	  if (fgets (buf1, sizeof (buf1) - 1, in) == NULL)
+	    log_msg ("Read error in line %d => Ignore line", line);
 	  break;
 	default:
-	  fgets (buf1, sizeof (buf1) - 1, in);
+	  if (fgets (buf1, sizeof (buf1) - 1, in) == NULL) {};
 	  log_msg ("Parse error in line %d: %c%s", line, c, buf1);
 	  break;
 	}
