@@ -1,10 +1,9 @@
-/* Copyright (c) 1996,1997, 1998, 1999, 2000, 2002, 2003, 2005 Thorsten Kukuk
+/* Copyright (c) 1996-2006, 2011, 2014 Thorsten Kukuk
    Author: Thorsten Kukuk <kukuk@suse.de>
 
    The YP Server is free software; you can redistribute it and/or
-   modify it under the terms of the GNU General Public License as
-   published by the Free Software Foundation; either version 2 of the
-   License, or (at your option) any later version.
+   modify it under the terms of the GNU General Public License
+   version 2 as published by the Free Software Foundation.
 
    The YP Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -13,32 +12,24 @@
 
    You should have received a copy of the GNU General Public
    License along with the YP Server; see the file COPYING. If
-   not, write to the Free Software Foundation, Inc., 675 Mass Ave,
-   Cambridge, MA 02139, USA. */
-
-#define _GNU_SOURCE
+   not, write to the Free Software Foundation, Inc., 51 Franklin Street,
+   Suite 500, Boston, MA 02110-1335, USA. */
 
 #if defined(HAVE_CONFIG_H)
 #include "config.h"
 #endif
 
-#ifdef HAVE_ALLOCA_H
 #include <alloca.h>
-#endif /* HAVE_ALLOCA_H */
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#ifdef HAVE_GETOPT_H
 #include <getopt.h>
-#endif /* HAVE_GETOPT_H */
 #include <fcntl.h>
 #include <ctype.h>
 #include <netdb.h>
 #include <rpc/rpc.h>
-
-#include "yp.h"
-#include "compat.h"
+#include <rpcsvc/yp_prot.h>
 
 #if defined (__NetBSD__) || (defined(__GLIBC__) && (__GLIBC__ == 2 && __GLIBC_MINOR__ == 0))
 /* <rpc/rpc.h> is missing the prototype */
@@ -47,10 +38,15 @@ int callrpc (char *host, u_long prognum, u_long versnum, u_long procnum,
 #endif
 #include <sys/param.h>
 #include <sys/time.h>
+#include <sys/stat.h>
+
+#if defined(HAVE_COMPAT_LIBGDBM)
 
 #if defined(HAVE_LIBGDBM)
-
 #include <gdbm.h>
+#elif defined(HAVE_LIBQDBM)
+#include <hovel.h>
+#endif
 
 #define ypdb_store gdbm_store
 #define YPDB_REPLACE GDBM_REPLACE
@@ -66,6 +62,30 @@ static GDBM_FILE dbm;
 #define ypdb_close dbm_close
 static DBM *dbm;
 
+#elif defined (HAVE_LIBTC)
+
+#include <tcbdb.h>
+
+#define YPDB_REPLACE 1
+
+static TCBDB *dbm;
+
+static inline int
+ypdb_store(TCBDB *dbm, datum key, datum data, int mode)
+{
+  if (mode != YPDB_REPLACE)
+    return 1;
+
+  return !tcbdbput(dbm, key.dptr, key.dsize, data.dptr, data.dsize);
+}
+
+static inline void
+ypdb_close (TCBDB *dbm)
+{
+  tcbdbclose (dbm);
+  tcbdbdel (dbm);
+}
+
 #else
 
 #error "No database found or selected!"
@@ -123,10 +143,17 @@ create_file (char *fileName, char *dbmName, char *masterName,
 
   filename = calloc (1, strlen (dbmName) + 3);
   sprintf (filename, "%s~", dbmName);
-#if defined(HAVE_LIBGDBM)
+#if defined(HAVE_COMPAT_LIBGDBM)
   dbm = gdbm_open (filename, 0, GDBM_NEWDB | GDBM_FAST, 0600, NULL);
 #elif defined(HAVE_NDBM)
   dbm = dbm_open (filename, O_CREAT | O_RDWR, 0600);
+#elif defined(HAVE_LIBTC)
+  dbm = tcbdbnew();
+  if (!tcbdbopen(dbm, filename, BDBOWRITER | BDBOCREAT))
+  {
+    tcbdbdel(dbm);
+    dbm = NULL;
+  }
 #endif
   if (dbm == NULL)
     {
@@ -209,25 +236,7 @@ create_file (char *fileName, char *dbmName, char *masterName,
     {
       char *cptr;
 
-#ifdef HAVE_GETLINE
       ssize_t n = getline (&key, &keylen, input);
-#elif HAVE_GETDELIM
-      ssize_t n = getdelim (&key, &keylen, '\n', input);
-#else
-      ssize_t n;
-
-      if (key == NULL)
-	{
-	  keylen = 8096;
-	  key = malloc (keylen);
-	}
-      key[0] = '\0';
-      fgets (key, keylen - 1, input);
-      if (key != NULL)
-	n = strlen (key);
-      else
-	n = 0;
-#endif
       if (n < 1)
 	break;
       if (key[n - 1] == '\n' || key[n - 1] == '\r')
@@ -265,16 +274,8 @@ create_file (char *fileName, char *dbmName, char *masterName,
 	    {
 	      char *nkey = NULL;
 	      size_t nkeylen = 0;
-#ifdef HAVE_GETLINE
-	      getline (&nkey, &nkeylen, input);
-#elif HAVE_GETDELIM
-	      getdelim (&nkey, &nkeylen, '\n', input);
-#else
-	      nkeylen = 8096;
-	      nkey = malloc (nkeylen);
-	      nkey[0] = '\0';
-	      fgets (nkey, nkeylen - 1, input);
-#endif
+	      if (getline (&nkey, &nkeylen, input) == -1)
+		break;
 
 	      cptr = nkey;
 	      while ((*cptr == ' ') || (*cptr == '\t'))
@@ -309,22 +310,8 @@ create_file (char *fileName, char *dbmName, char *masterName,
 	  {
 	    char *nkey;
 	    size_t nkeylen = 0;
-#ifdef HAVE_GETLINE
 	    ssize_t n = getline (&nkey, &nkeylen, input);
-#elif HAVE_GETDELIM
-	    ssize_t n = getdelim (&nkey, &nkeylen, '\n', input);
-#else
-	    ssize_t n;
 
-	    nkeylen = 8096;
-	    nkey = malloc (nkeylen);
-	    nkey[0] = '\0';
-	    fgets (nkey, nkeylen - 1, input);
-	    if (nkey != NULL)
-	      n = strlen (nkey);
-	    else
-	      n = 0;
-#endif
 	    if (n < 1)
 	      break;
 	    if (nkey[n - 1] == '\n' || nkey[n - 1] == '\r')
@@ -455,26 +442,40 @@ create_file (char *fileName, char *dbmName, char *masterName,
 #endif
 #else
   unlink (dbmName);
+#if defined(HAVE_LIBTC)
+	  chmod(filename, S_IRUSR|S_IWUSR);
+#endif
   rename (filename, dbmName);
 #endif
   free (filename);
+
+  if (strcmp (fileName, "-") != 0)
+    fclose (input);
 }
 
 static void
 dump_file (char *dbmName)
 {
   datum key, data;
-#if defined(HAVE_LIBGDBM)
+#if defined(HAVE_COMPAT_LIBGDBM)
   dbm = gdbm_open (dbmName, 0, GDBM_READER, 0600, NULL);
 #elif defined(HAVE_NDBM)
   dbm = dbm_open (dbmName, O_RDONLY, 0600);
+#elif defined(HAVE_LIBTC)
+  dbm = tcbdbnew();
+  if (!tcbdbopen (dbm, dbmName, BDBOREADER | BDBONOLCK))
+  {
+    tcbdbdel(dbm);
+    dbm = NULL;
+  }
 #endif
   if (dbm == NULL)
     {
       fprintf (stderr, "makedbm: Cannot open %s\n", dbmName);
+      fprintf (stderr, "makedbm: Consider rebuilding maps using ypinit\n");
       exit (1);
     }
-#if defined(HAVE_LIBGDBM)
+#if defined(HAVE_COMPAT_LIBGDBM)
   for (key = gdbm_firstkey (dbm); key.dptr; key = gdbm_nextkey (dbm, key))
     {
       data = gdbm_fetch (dbm, key);
@@ -505,6 +506,32 @@ dump_file (char *dbmName)
 	      data.dsize, data.dptr);
       key = dbm_nextkey (dbm);
     }
+#elif defined(HAVE_LIBTC)
+  {
+    BDBCUR *cur;
+    cur = tcbdbcurnew (dbm);
+    if (tcbdbcurfirst (cur))
+      {
+        while ((key.dptr = tcbdbcurkey (cur, &key.dsize)) != NULL)
+          {
+            data.dptr = tcbdbcurval (cur, &data.dsize);
+            if (!data.dptr)
+	      {
+	        fprintf (stderr, "Error:\n");
+	        perror (dbmName);
+	        exit (1);
+              }
+
+            printf ("%.*s\t%.*s\n",
+                  key.dsize, key.dptr,
+                  data.dsize, data.dptr);
+
+            if (!tcbdbcurnext (cur))
+              break;
+          }
+      }
+    tcbdbcurdel (cur);
+  }
 #endif
   ypdb_close (dbm);
 }
@@ -524,6 +551,61 @@ send_clear (void)
     }
 }
 
+static char *
+get_canonical_hostname (const char *hostname)
+{
+#if USE_FQDN
+  struct addrinfo hints, *res0, *res1;
+  int error;
+  char *host = NULL;
+
+  memset (&hints, 0, sizeof(struct addrinfo));
+  hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
+#if 0
+  hints.ai_flags = AI_CANONNAME;  /* get the official name of the host */
+#endif
+
+  if ((error = getaddrinfo (hostname, NULL, &hints, &res0)))
+    {
+#if 0
+      printf ("getaddrinfo: %s\n", gai_strerror (error));
+#endif
+      return strdup (hostname);
+    }
+
+  res1 = res0;
+
+  while (res1)
+    {
+      char hostbuf[NI_MAXHOST];
+
+      if ((error = getnameinfo (res1->ai_addr, res1->ai_addrlen,
+                                (char *)&hostbuf, sizeof (hostbuf),
+                                NULL, 0, NI_NAMEREQD)) == 0)
+        {
+          host = strdup (hostbuf);
+          break;
+        }
+#if 0
+      else
+        printf ("getnameinfo: %s\n", gai_strerror (error));
+#endif
+
+      res1 = res1->ai_next;
+    }
+
+  if (host == NULL)
+    host = strdup (res0->ai_canonname);
+
+  freeaddrinfo (res0);
+
+  return host;
+#else
+  return strdup (hostname);
+#endif
+}
+
+
 static void
 Usage (int exit_code)
 {
@@ -654,23 +736,15 @@ main (int argc, char *argv[])
 	{
 	  if (strlen (masterName) == 0)
 	    {
+	      char *cp;
+
 	      if (gethostname (masterName, sizeof (masterName)) < 0)
 		perror ("gethostname");
-#if USE_FQDN
-	      else
-		{
-		  struct hostent *hp;
 
-		  if (!(hp = gethostbyname (masterName)))
-		    perror ("gethostbyname()");
-		  else
-		    {
-		      strncpy (masterName, hp->h_name, MAXHOSTNAMELEN);
-		      masterName[MAXHOSTNAMELEN] = '\0';
-		    }
-		}
-#endif
+	      cp = get_canonical_hostname (masterName);
+	      strncpy (masterName, cp, sizeof (masterName) -1);
 	    }
+
 	  create_file (argv[0], argv[1], masterName, domainName,
 		       inputName, outputName, aliases, shortline,
 		       b_flag, s_flag, remove_comments, check_limit);
