@@ -1,13 +1,12 @@
 /*
-   Copyright (c) 1996, 1997, 1998, 1999, 2001, 2005 Thorsten Kukuk, <kukuk@suse.de>
+   Copyright (c) 1996-2006, 2010, 2011, 2012, 2014, 2015, 2016 Thorsten Kukuk, <kukuk@thkukuk.de>
    Copyright (c) 1994, 1995, 1996 Olaf Kirch, <okir@monad.swb.de>
 
    This file is part of the NYS YP Server.
 
-   The NYS YP Server is free software; you can redistribute it and/or
-   modify it under the terms of the GNU General Public License as
-   published by the Free Software Foundation; either version 2 of the
-   License, or (at your option) any later version.
+   The YP Server is free software; you can redistribute it and/or
+   modify it under the terms of the GNU General Public License
+   version 2 as published by the Free Software Foundation.
 
    The NYS YP Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -16,8 +15,8 @@
 
    You should have received a copy of the GNU General Public
    License along with the NYS YP Server; see the file COPYING.  If
-   not, write to the Free Software Foundation, Inc., 675 Mass Ave,
-   Cambridge, MA 02139, USA. */
+   not, write to the Free Software Foundation, Inc., 51 Franklin Street,
+   Suite 500, Boston, MA 02110-1335, USA. */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -37,25 +36,18 @@
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
+#include <arpa/inet.h>
 #include <rpc/rpc.h>
-#include <rpc/pmap_clnt.h>
-#if defined(HAVE_RPC_SVC_SOC_H)
-#include <rpc/svc_soc.h>
-#endif
-#include "yppasswd.h"
-#if defined(HAVE_GETOPT_H)
+#include <rpc/nettype.h>
+#include <rpcsvc/yp_prot.h>
+#include <rpcsvc/yppasswd.h>
 #include <getopt.h>
-#endif
 
+#include "yppwd_local.h"
 #include "log_msg.h"
-#include "compat.h"
+#include "pidfile.h"
+#include "access.h"
 
-#ifdef HAVE_PATHS_H
-#include <paths.h>
-#endif
-#ifndef _PATH_VARRUN
-#define _PATH_VARRUN "/etc/"
-#endif
 #define _YPPASSWDD_PIDFILE _PATH_VARRUN"yppasswdd.pid"
 
 int use_shadow = 0;
@@ -64,8 +56,8 @@ int allow_chfn = 0;
 int solaris_mode = -1;
 int x_flag = -1;
 
-#define xprt_addr(xprt)	(svc_getcaller(xprt)->sin_addr)
-#define xprt_port(xprt)	ntohs(svc_getcaller(xprt)->sin_port)
+static int foreground_flag = 0;
+
 void yppasswdprog_1 (struct svc_req *rqstp, SVCXPRT * transp);
 void reaper (int sig);
 
@@ -97,7 +89,20 @@ yppasswdprog_1 (struct svc_req *rqstp, SVCXPRT * transp)
   memset ((char *) &argument, 0, sizeof (argument));
   if (!svc_getargs (transp, xdr_argument, (caddr_t) &argument))
     {
+      char namebuf6[INET6_ADDRSTRLEN];
+      struct netconfig *nconf;
+      const struct netbuf *rqhost = svc_getrpccaller (rqstp->rq_xprt);
+
+      nconf = getnetconfigent (rqstp->rq_xprt->xp_netid);
+
+      log_msg ("cannot decode arguments for %d from %s",
+	       rqstp->rq_proc, taddr2ipstr (nconf, rqhost,
+					    namebuf6, sizeof (namebuf6)));
+      /* try to free already allocated memory during decoding */
+      svc_freeargs (transp, xdr_argument, (caddr_t) &argument);
       svcerr_decode (transp);
+      freenetconfigent (nconf);
+
       return;
     }
   result = yppasswdproc_pwupdate_1 (&argument, rqstp);
@@ -113,76 +118,12 @@ yppasswdprog_1 (struct svc_req *rqstp, SVCXPRT * transp)
     }
 }
 
-/* Create a pidfile on startup */
-static void
-create_pidfile (void)
-{
-  int fd, left, written;
-  pid_t pid;
-  char pbuf[50], *ptr;
-  struct flock lock;
-
-  fd = open (_YPPASSWDD_PIDFILE, O_CREAT | O_RDWR,
-	     S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
-  if (fd < 0)
-    {
-      log_msg ("cannot create pidfile %s", _YPPASSWDD_PIDFILE);
-      if (debug_flag)
-	log_msg ("\n");
-    }
-
-  lock.l_type = F_WRLCK;
-  lock.l_start = 0;
-  lock.l_whence = SEEK_SET;
-  lock.l_len = 0;
-
-  /* Is the pidfile locked by another ypserv ? */
-  if (fcntl (fd, F_GETLK, &lock) < 0)
-    {
-      log_msg ("fcntl error");
-      if (debug_flag)
-	log_msg ("\n");
-    }
-  if (lock.l_type == F_UNLCK)
-    pid = 0;	        /* false, region is not locked by another proc */
-  else
-    pid = lock.l_pid;	/* true, return pid of lock owner */
-
-  if (0 != pid)
-    {
-      log_msg ("rpc.yppasswdd already running (pid %d) - exiting", pid);
-      if (debug_flag)
-	log_msg ("\n");
-      exit (1);
-    }
-
-  /* write lock */
-  lock.l_type = F_WRLCK;
-  lock.l_start = 0;
-  lock.l_whence = SEEK_SET;
-  lock.l_len = 0;
-  if (0 != fcntl (fd, F_SETLK, &lock))
-    log_msg ("cannot lock pidfile");
-  sprintf (pbuf, "%ld\n", (long) getpid ());
-  left = strlen (pbuf);
-  ptr = pbuf;
-  while (left > 0)
-    {
-      if ((written = write (fd, ptr, left)) <= 0)
-	return;			/* error */
-      left -= written;
-      ptr += written;
-    }
-  return;
-}
-
-
 static void
 usage (FILE * fp, int n)
 {
-  fputs ("Usage: rpc.yppasswdd [--debug] [-s shadowfile] [-p passwdfile] [-e chsh|chfn]\n", fp);
-  fputs ("       rpc.yppasswdd [--debug] [-D directory] [-e chsh|chfn]\n", fp);
-  fputs ("       rpc.yppasswdd [--debug] [-x program |-E program] [-e chsh|chfn]\n", fp);
+  fputs ("Usage: rpc.yppasswdd [--debug] [-s shadowfile] [-p passwdfile] [-e chsh|chfn] [-f|--foreground]\n", fp);
+  fputs ("       rpc.yppasswdd [--debug] [-D directory] [-e chsh|chfn] [-f|--foreground]\n", fp);
+  fputs ("       rpc.yppasswdd [--debug] [-x program |-E program] [-e chsh|chfn] [-f|--foreground]\n", fp);
   fputs ("       rpc.yppasswdd --port number\n", fp);
   fputs ("       rpc.yppasswdd --version\n", fp);
   exit (n);
@@ -202,7 +143,7 @@ sig_child (int sig UNUSED)
 static void
 sig_quit (int sig UNUSED)
 {
-  pmap_unset (YPPASSWDPROG, YPPASSWDVERS);
+  rpcb_unset (YPPASSWDPROG, YPPASSWDVERS, NULL);
   unlink (_YPPASSWDD_PIDFILE);
   exit (0);
 }
@@ -253,9 +194,11 @@ install_sighandler (void)
 int
 main (int argc, char **argv)
 {
-  SVCXPRT *transp;
-  int my_port = -1, my_socket;
+  struct netconfig *nconf;
+  int my_port = -1;
+  void *nc_handle;
   int c;
+  int could_register = 0;
 
   /* Initialize logging. */
   openlog ("rpc.yppasswdd", LOG_PID, LOG_AUTH);
@@ -266,16 +209,17 @@ main (int argc, char **argv)
       int option_index = 0;
       static struct option long_options[] =
       {
-	{"version", no_argument, NULL, '\255'},
+	{"version", no_argument, NULL, 'v'},
 	{"usage", no_argument, NULL, 'h'},
 	{"help", no_argument, NULL, 'h'},
 	{"execute", required_argument, NULL, 'x'},
+	{"foreground", no_argument, NULL, 'f'},
 	{"debug", no_argument, NULL, '\254'},
 	{"port", required_argument, NULL, '\253'},
 	{NULL, 0, NULL, '\0'}
       };
 
-      c=getopt_long (argc, argv, "e:p:s:uhvD:E:x:m", long_options,
+      c=getopt_long (argc, argv, "e:p:s:fuhvD:E:x:m", long_options,
 		     &option_index);
       if (c == EOF)
 	break;
@@ -295,13 +239,14 @@ main (int argc, char **argv)
 	  solaris_mode = 0;
 	  path_passwd = optarg;
 	  break;
+	case 'f':
+	  foreground_flag = 1;
+	  break;
 	case 's':
 	  if (solaris_mode == 1)
 	    usage (stderr, 1);
 	  solaris_mode = 0;
-#ifdef HAVE_GETSPNAM
 	  path_shadow = optarg;
-#endif
 	  break;
 	case 'D':
 	  if (solaris_mode == 0)
@@ -309,10 +254,8 @@ main (int argc, char **argv)
 	  solaris_mode = 1;
 	  path_passwd = malloc (strlen (optarg) + 8);
 	  sprintf (path_passwd, "%s/passwd", optarg);
-#ifdef HAVE_GETSPNAM
 	  path_shadow = malloc (strlen (optarg) + 8);
 	  sprintf (path_shadow, "%s/shadow", optarg);
-#endif
 	  break;
 	case 'E':
 	  external_update_program = strdup(optarg);
@@ -334,10 +277,16 @@ main (int argc, char **argv)
 	  break;
 	case '\253':
           my_port = atoi (optarg);
+	  if (my_port <= 0 || my_port > 0xffff) {
+		/* Invalid port number */
+	    fprintf (stdout, "Warning: rpc.yppasswdd: Invalid port %d (0x%x)\n",
+			my_port, my_port);
+		my_port = -1;
+	  }
           if (debug_flag)
             log_msg ("Using port %d\n", my_port);
           break;
-	case '\255':
+	case 'v':
 #if CHECKROOT
 	  fprintf (stdout, "rpc.yppasswdd - YP server version %s (with CHECKROOT)\n",
 		   VERSION);
@@ -373,7 +322,6 @@ main (int argc, char **argv)
       exit (-1);
     }
   sprintf (path_passwd_old, "%s.OLD", path_passwd);
-#ifdef HAVE_GETSPNAM
   /* Create tmp and .OLD file names for "shadow" */
   path_shadow_tmp = malloc (strlen (path_shadow) + 5);
   if (path_shadow_tmp == NULL)
@@ -389,7 +337,6 @@ main (int argc, char **argv)
       exit (-1);
     }
   sprintf (path_shadow_old, "%s.OLD", path_shadow);
-#endif /* HAVE_GETSPNAM */
 
   if (debug_flag)
     {
@@ -400,7 +347,7 @@ main (int argc, char **argv)
       log_msg ("rpc.yppasswdd - NYS YP server version %s\n", VERSION);
 #endif /* CHECKROOT */
     }
-  else
+  else if (!foreground_flag)
     {
       int i;
 
@@ -425,70 +372,192 @@ main (int argc, char **argv)
 
       if (i < 0)
 	{
-	  log_msg ("rpc.yppasswdd: cannot fork: %s\n", strerror (errno));
-	  exit (-1);
+	  int err = errno;
+	  log_msg ("rpc.yppasswdd: cannot fork: %s\n", strerror (err));
+	  exit (err);
 	}
 
       for (i = 0; i < getdtablesize (); ++i)
         close (i);
       errno = 0;
 
-      chdir ("/");
+      if (chdir ("/") == -1)
+	{
+	  int err = errno;
+	  log_msg ("rpc.yppasswdd: chdir failed: %s\n", strerror (err));
+	  exit (err);
+	}
       umask(0);
       i = open("/dev/null", O_RDWR);
-      dup(i);
-      dup(i);
+      if (i == -1)
+	{
+	  int err = errno;
+	  log_msg ("rpc.yppasswdd: open /dev/null failed: %s\n",
+		   strerror (err));
+	  exit (err);
+	}
+
+      /* two dups, we have stdin, stdout, stderr */
+      if (dup(i) == -1)
+	{
+	  int err = errno;
+	  log_msg ("rpc.yppasswdd: dup failed: %s\n", strerror (err));
+	  exit (err);
+	}
+
+      if (dup(i) == -1)
+	{
+	  int err = errno;
+	  log_msg ("rpc.yppasswdd: dup failed: %s\n", strerror (err));
+	  exit (err);
+	}
     }
 
-  create_pidfile ();
+  create_pidfile (_YPPASSWDD_PIDFILE, "rpc.yppasswdd");
 
   /* Register a signal handler to reap children after they terminated */
   install_sighandler ();
 
   /* Create the RPC server */
-  pmap_unset (YPPASSWDPROG, YPPASSWDVERS);
+  rpcb_unset (YPPASSWDPROG, YPPASSWDVERS, NULL);
 
-  if (my_port >= 0)
+  nc_handle = __rpc_setconf ("netpath");   /* open netconfig file */
+  if (nc_handle == NULL)
     {
-      struct sockaddr_in s_in;
-      int result;
+      log_msg ("could not read /etc/netconfig, exiting..");
+      return 1;
+    }
 
-      my_socket = socket (AF_INET, SOCK_DGRAM, 0);
-      if (my_socket < 0)
+
+  while ((nconf = __rpc_getconf (nc_handle)))
+    {
+      SVCXPRT *xprt;
+      struct sockaddr *sa;
+      struct sockaddr_in sin;
+      struct sockaddr_in6 sin6;
+      int sock;
+      sa_family_t family; /* AF_INET, AF_INET6 */
+      int type; /* SOCK_DGRAM (udp), SOCK_STREAM (tcp) */
+      int proto; /* IPPROTO_UDP, IPPROTO_TCP */
+
+      if (debug_flag)
+        log_msg ("Register ypserv for %s,%s",
+                 nconf->nc_protofmly, nconf->nc_proto);
+
+      if (strcmp (nconf->nc_protofmly, "inet6") == 0)
+        family = AF_INET6;
+      else if (strcmp (nconf->nc_protofmly, "inet") == 0)
+        family = AF_INET;
+      else
+        continue; /* we don't support nconf->nc_protofmly */
+
+      if (strcmp (nconf->nc_proto, "udp") == 0)
         {
-          log_msg ("can not create UDP: %s", strerror (errno));
-          exit (1);
+          type = SOCK_DGRAM;
+          proto = IPPROTO_UDP;
+        }
+      else if (strcmp (nconf->nc_proto, "tcp") == 0)
+        {
+          type = SOCK_STREAM;
+          proto = IPPROTO_TCP;
+        }
+      else
+        continue; /* We don't support nconf->nc_proto */
+
+      if ((sock = socket (family, type, proto)) < 0)
+        {
+          log_msg ("Cannot create socket for %s,%s: %s",
+                   nconf->nc_protofmly, nconf->nc_proto,
+                   strerror (errno));
+          continue;
         }
 
-      memset ((char *) &s_in, 0, sizeof (s_in));
-      s_in.sin_family = AF_INET;
-      s_in.sin_addr.s_addr = htonl (INADDR_ANY);
-      s_in.sin_port = htons (my_port);
-
-      result = bind (my_socket, (struct sockaddr *) &s_in,
-                     sizeof (s_in));
-      if (result < 0)
+      if (family == AF_INET6)
         {
-          log_msg ("rpc.yppasswdd: can not bind UDP: %s ", strerror (errno));
-          exit (1);
-        }
-    }
-  else
-    my_socket = RPC_ANYSOCK;
+          /* Disallow v4-in-v6 to allow host-based access checks */
 
-  transp = svcudp_create (my_socket);
-  if (transp == NULL)
-    {
-      log_msg ("cannot create udp service.\n");
-      exit (1);
+          int i;
+
+          if (setsockopt (sock, IPPROTO_IPV6, IPV6_V6ONLY,
+                          &i, sizeof(i)) == -1)
+            {
+              log_msg ("ERROR: cannot disable v4-in-v6 on %s6 socket",
+                       nconf->nc_proto);
+              return 1;
+            }
+        }
+
+      switch (family)
+        {
+        case AF_INET:
+          memset (&sin, 0, sizeof(sin));
+          sin.sin_family = AF_INET;
+          if (my_port > 0)
+            sin.sin_port = htons (my_port);
+          sa = (struct sockaddr *)(void *)&sin;
+          break;
+        case AF_INET6:
+          memset (&sin6, 0, sizeof (sin6));
+          sin6.sin6_family = AF_INET6;
+          if (my_port > 0)
+            sin6.sin6_port = htons (my_port);
+          sa = (struct sockaddr *)(void *)&sin6;
+          break;
+        default:
+          log_msg ("Unsupported address family %d", family);
+          return -1;
+        }
+
+      if (bindresvport_sa (sock, sa) == -1)
+        {
+          if (my_port > 0)
+            log_msg ("Cannot bind to reserved port %d (%s)",
+                     my_port, strerror (errno));
+          else
+            log_msg ("bindresvport failed: %s",
+                     strerror (errno));
+          return 1;
+        }
+
+      if (type == SOCK_STREAM)
+        {
+          listen (sock, SOMAXCONN);
+          xprt = svc_vc_create (sock, 0, 0);
+        }
+      else
+        xprt = svc_dg_create (sock, 0, 0);
+
+      if (xprt == NULL)
+        {
+          log_msg ("terminating: cannot create rpcbind handle");
+          return 1;
+        }
+
+      rpcb_unset (YPPASSWDPROG, YPPASSWDVERS, nconf);
+      if (!svc_reg (xprt, YPPASSWDPROG, YPPASSWDVERS, yppasswdprog_1, nconf))
+        {
+          log_msg ("unable to register (YPPASSWDPROG, 1) for %s, %s.",
+                   nconf->nc_protofmly, nconf->nc_proto);
+          continue;
+        }
+      else
+	could_register = 1;
     }
-  if (!svc_register (transp, YPPASSWDPROG, YPPASSWDVERS, yppasswdprog_1,
-		     IPPROTO_UDP))
+  __rpc_endconf (nc_handle);
+
+  if (!could_register)
     {
-      log_msg ("unable to register yppaswdd udp service.\n");
-      exit (1);
+      log_msg ("terminating: rpcbind not running?");
+      return 1;
     }
 
+  /* If we use systemd as an init system, we may want to give it
+     a message, that this daemon is ready to accept connections.
+     At this time, sockets for receiving connections are already
+     created, so we can say we're ready now. It is a nop if we
+     don't use systemd. */
+  announce_ready();
+
   /* Run the server */
   svc_run ();
   log_msg ("svc_run returned\n");
