@@ -1,4 +1,4 @@
-/* Copyright (c) 1996-2004 Thorsten Kukuk
+/* Copyright (c) 1996-2005, 2014, 2015, 2016 Thorsten Kukuk
    Author: Thorsten Kukuk <kukuk@suse.de>
 
    The YP Server is free software; you can redistribute it and/or
@@ -12,15 +12,13 @@
 
    You should have received a copy of the GNU General Public
    License along with the YP Server; see the file COPYING. If
-   not, write to the Free Software Foundation, Inc., 675 Mass Ave,
-   Cambridge, MA 02139, USA. */
+   not, write to the Free Software Foundation, Inc., 51 Franklin Street,
+   Suite 500, Boston, MA 02110-1335, USA. */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#define _GNU_SOURCE
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -42,21 +40,18 @@
 #include <memory.h>
 #if defined(HAVE_LIBGDBM)
 #include <gdbm.h>
+#elif defined(HAVE_LIBQDBM)
+#include <hovel.h>
 #elif defined(HAVE_NDBM)
 #include <ndbm.h>
 #include <fcntl.h>
+#elif defined(HAVE_LIBTC)
+#include <tcbdb.h>
 #endif
-#if defined(HAVE_GETOPT_H)
 #include <getopt.h>
-#endif /* HAVE_GETOPT_H */
-#include "compat.h"
 
 #include "log_msg.h"
 
-#ifndef YPMAPDIR
-#define YPMAPDIR "/var/yp"
-#endif
-
 struct hostlist {
   char *hostname;
   struct hostlist *next;
@@ -113,6 +108,8 @@ yppush_err_string (enum yppush_status status)
       return "ypxfr error";
     case YPPUSH_REFUSED:
       return "Transfer request refused by ypserv";
+    case YPPUSH_NOALIAS:
+      return "Alias not found for map or domain";
     }
   return "YPPUSH: Unknown Error, this should not happen!";
 }
@@ -135,27 +132,28 @@ bool_t
 yppushproc_xfrresp_1_svc (yppushresp_xfr *req,
 			  void *resp UNUSED, struct svc_req *rqstp)
 {
-  struct sockaddr_in *sin;
-  char *h;
-  struct hostent *hp;
+  char hostbuf[NI_MAXHOST];
+  struct netconfig *nconf;
+  struct netbuf *nbuf;
 
   if (verbose_flag > 1)
     log_msg ("yppushproc_xfrresp_1_svc");
 
-  sin = svc_getcaller (rqstp->rq_xprt);
-
-  hp = gethostbyaddr ((char *) &sin->sin_addr.s_addr,
-		      sizeof (sin->sin_addr.s_addr), AF_INET);
-  h = (hp && hp->h_name) ? hp->h_name : inet_ntoa (sin->sin_addr);
+  nbuf = svc_getrpccaller (rqstp->rq_xprt);
+  nconf = getnetconfigent (rqstp->rq_xprt->xp_netid);
 
   if (verbose_flag)
     {
-      log_msg ("Status received from ypxfr on %s:", h);
-      log_msg ("\tTransfer %sdone: %s", req->status == YPPUSH_SUCC ? "" : "not ",
-	      yppush_err_string (req->status));
+      log_msg ("Status received from ypxfr on %s",
+	       taddr2host (nconf, nbuf, hostbuf, sizeof (hostbuf)));
+      log_msg ("\tTransfer %sdone: %s",
+	       req->status == YPPUSH_SUCC ? "" : "not ",
+	       yppush_err_string (req->status));
     }
   else if (req->status != YPPUSH_SUCC)
-    log_msg ("%s: %s", h, yppush_err_string (req->status));
+    log_msg ("%s: %s", taddr2host (nconf, nbuf, hostbuf, sizeof (hostbuf)),
+	     yppush_err_string (req->status));
+  freenetconfigent (nconf);
 
   return TRUE;
 }
@@ -193,6 +191,18 @@ yppush_xfrrespprog_1(struct svc_req *rqstp, register SVCXPRT *transp)
   memset ((char *)&argument, 0, sizeof (argument));
   if (!svc_getargs (transp, _xdr_argument, (caddr_t) &argument))
     {
+      char hostbuf[NI_MAXHOST];
+      struct netconfig *nconf;
+      struct netbuf *nbuf = svc_getrpccaller (rqstp->rq_xprt);
+
+      nconf = getnetconfigent (rqstp->rq_xprt->xp_netid);
+
+      log_msg ("cannot decode arguments for %d from %s",
+	       rqstp->rq_proc,
+	       taddr2host (nconf, nbuf, hostbuf, sizeof (hostbuf)));
+      /* try to free already allocated memory during decoding */
+      svc_freeargs (transp, _xdr_argument, (caddr_t) &argument);
+      freenetconfigent (nconf);
       svcerr_decode (transp);
       return;
     }
@@ -206,11 +216,6 @@ yppush_xfrrespprog_1(struct svc_req *rqstp, register SVCXPRT *transp)
     exit (1);
   }
 
-#if 0
-  /* XXX */
-  if (!yppush_xfrrespprog_1_freeresult (transp, _xdr_result, (caddr_t) &result))
-    log_msg ("unable to free results");
-#endif
   if (rqstp->rq_proc != YPPUSHPROC_NULL)
     exit (0);
 
@@ -231,7 +236,7 @@ yppush_svc_run (char *target)
     {
       readfds = svc_fdset;
 
-      switch (select (_rpc_dtablesize (), &readfds, (void *) 0, (void *) 0, &tr))
+      switch (select (svc_maxfd+1, &readfds, (void *) 0, (void *) 0, &tr))
 	{
 	case -1:
 	  if (errno == EINTR)
@@ -239,7 +244,7 @@ yppush_svc_run (char *target)
 	      tr = tb;		/* Read the Linux select.2 manpage ! */
 	      continue;
 	    }
-	  log_msg ("svc_run: - select failed (%s)", strerror (errno));
+	  log_msg ("yppush_svc_run: - select failed (%s)", strerror (errno));
 	  return;
 	case 0:
 	  log_msg ("%s->%s: Callback timed out", current_map, target);
@@ -257,10 +262,12 @@ get_dbm_entry (char *key)
   static char mappath[MAXPATHLEN + 2];
   char *val;
   datum dkey, dval;
-#if defined(HAVE_LIBGDBM)
+#if defined(HAVE_COMPAT_LIBGDBM)
   GDBM_FILE dbm;
 #elif defined (HAVE_NDBM)
   DBM *dbm;
+#elif defined (HAVE_LIBTC)
+  TCBDB *dbm;
 #endif
 
   if (strlen (YPMAPDIR) + strlen (DomainName) + strlen (current_map) + 3 < MAXPATHLEN)
@@ -271,23 +278,33 @@ get_dbm_entry (char *key)
       exit (1);
     }
 
-#if defined(HAVE_LIBGDBM)
+#if defined(HAVE_COMPAT_LIBGDBM)
   dbm = gdbm_open (mappath, 0, GDBM_READER, 0600, NULL);
 #elif defined(HAVE_NDBM)
-  dbm = dbm_open (mappath, O_CREAT | O_RDWR, 0600);
+  dbm = dbm_open (mappath, O_RDONLY, 0600);
+#elif defined(HAVE_LIBTC)
+  dbm = tcbdbnew();
+  if (!tcbdbopen(dbm, mappath, BDBOREADER | BDBONOLCK))
+    {
+      tcbdbdel(dbm);
+      dbm = NULL;
+    }
 #endif
   if (dbm == NULL)
     {
       log_msg ("YPPUSH: Cannot open %s", mappath);
+      log_msg ("YPPUSH: consider rebuilding maps using ypinit");
       exit (1);
     }
 
   dkey.dptr = key;
   dkey.dsize = strlen (dkey.dptr);
-#if defined(HAVE_LIBGDBM)
+#if defined(HAVE_COMPAT_LIBGDBM)
   dval = gdbm_fetch (dbm, dkey);
 #elif defined(HAVE_NDBM)
   dval = dbm_fetch (dbm, dkey);
+#elif defined(HAVE_LIBTC)
+  dval.dptr = tcbdbget (dbm, dkey.dptr, dkey.dsize, &dval.dsize);
 #endif
   if (dval.dptr == NULL)
     val = NULL;
@@ -297,10 +314,13 @@ get_dbm_entry (char *key)
       strncpy (val, dval.dptr, dval.dsize);
       val[dval.dsize] = 0;
     }
-#if defined(HAVE_LIBGDBM)
+#if defined(HAVE_COMPAT_LIBGDBM)
   gdbm_close (dbm);
 #elif defined(HAVE_NDBM)
   dbm_close (dbm);
+#elif defined(HAVE_LIBTC)
+  tcbdbclose (dbm);
+  tcbdbdel (dbm);
 #endif
   return val;
 }
@@ -397,7 +417,7 @@ static void
 child_sig_int (int sig UNUSED)
 {
   if (CallbackProg != 0)
-    svc_unregister (CallbackProg, 1);
+    svc_unreg (CallbackProg, 1);
   exit (1);
 }
 
@@ -405,82 +425,148 @@ static int
 yppush_foreach (const char *host)
 {
   SVCXPRT *CallbackXprt;
-  CLIENT *PushClient = NULL;
-  struct ypreq_xfr req;
+  CLIENT *PushClient;
+  struct ypreq_newxfr newreq;
   struct timeval tv = {10, 0};
   u_int transid;
   char server[YPMAXPEER + 2];
-  int sock;
-  struct sockaddr_in s_in;
-  struct sigaction sa;
+  int i, sock;
+  struct sigaction sig;
+  struct netconfig *nconf;
+  struct sockaddr *sa;
+  struct sockaddr_in sin;
+  struct sockaddr_in6 sin6;
+  enum clnt_stat res;
 
   if (verbose_flag > 1)
     log_msg ("yppush_foreach: host=%s", host);
 
-  sa.sa_handler = child_sig_int;
-  sigemptyset (&sa.sa_mask);
-#if defined(linux) || (defined(sun) && defined(__srv4__))
-  sa.sa_flags = SA_NOMASK;
+  sig.sa_handler = child_sig_int;
+  sigemptyset (&sig.sa_mask);
+  sig.sa_flags = SA_NOMASK;
   /* Do  not  prevent  the  signal   from   being
      received from within its own signal handler. */
-#endif
-  sigaction (SIGINT, &sa, NULL);
+  sigaction (SIGINT, &sig, NULL);
 
   if (strlen (host) < YPMAXPEER)
     sprintf (server, "%s", host);
   else
     {
-      log_msg ("YPPUSH ERROR: yppush_foreach: %.*s to long", host);
+      log_msg ("YPPUSH ERROR: yppush_foreach: %s to long", host);
       exit (1);
     }
 
-  PushClient = clnt_create (server, YPPROG, YPVERS, "udp");
+  PushClient = clnt_create (server, YPPROG, YPVERS, "datagram_n");
   if (PushClient == NULL)
     {
       clnt_pcreateerror (server);
       return 1;
     }
 
-
-  if (my_port >= 0)
+  /* Register a socket for IPv4 and, if supported, for IPv6, too */
+  if ((sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
     {
-      sock = socket (AF_INET, SOCK_DGRAM, 0);
-      if (sock < 0)
-        {
-          log_msg ("can not create UDP: %s", strerror (errno));
-          exit (1);
-        }
-      else
-        {
-          /* bind to socket */
-          memset ((char *) &s_in, 0, sizeof (s_in));
-          s_in.sin_family = AF_INET;
-          s_in.sin_addr.s_addr = htonl (INADDR_ANY);
-          s_in.sin_port = htons (my_port);
-
-          if  (bind (sock, (struct sockaddr *) &s_in, sizeof (s_in)) < 0)
-            {
-          log_msg ("yppush: can not bind UDP: %s ", strerror (errno));
-          exit (1);
-            }
-        }
-    }
-  else
-    sock = RPC_ANYSOCK;
-
-
-  CallbackXprt = svcudp_create (sock);
-  if (CallbackXprt == NULL)
-    {
-      log_msg ("YPPUSH: Cannot create callback transport to host \"%s\".", server);
+      log_msg ("Cannot create UDP socket for AF_INET: %s",
+	       strerror (errno));
       return 1;
     }
+
+  memset (&sin, 0, sizeof(sin));
+  sin.sin_family = AF_INET;
+  if (my_port > 0)
+    sin.sin_port = htons (my_port);
+  sa = (struct sockaddr *)(void *)&sin;
+
+  if (bindresvport_sa (sock, sa) == -1)
+    {
+      if (my_port > 0)
+	log_msg ("Cannot bind to reserved port %d (%s)",
+		 my_port, strerror (errno));
+      else
+	log_msg ("bindresvport failed: %s",
+		 strerror (errno));
+      return 1;
+    }
+
+  if ((CallbackXprt = svc_dg_create (sock, 0, 0)) == NULL)
+    {
+      log_msg ("terminating: cannot create rpcbind handle");
+      return 1;
+    }
+
+  nconf = getnetconfigent ("udp");
+  if (nconf == NULL)
+    {
+      log_msg ("YPPUSH: getnetconfigent (\"udp\") failed.");
+      exit (1);
+    }
   for (CallbackProg = 0x40000000; CallbackProg < 0x5fffffff; CallbackProg++)
     {
-      if (svc_register (CallbackXprt, CallbackProg, 1,
-			yppush_xfrrespprog_1, IPPROTO_UDP))
+      if (svc_reg (CallbackXprt, CallbackProg, 1,
+		   yppush_xfrrespprog_1, nconf))
 	break;
     }
+  freenetconfigent (nconf);
+
+  if (CallbackProg == 0x5FFFFFFF)
+    {
+      log_msg ("can't register yppush_xfrrespprog_1");
+      exit (1);
+    }
+  else if (verbose_flag > 1)
+    log_msg ("yppush_xfrrespprog_1 registered at %x", CallbackProg);
+
+  /* And now do the same for IPv6 */
+  if ((sock = socket (AF_INET6, SOCK_DGRAM, IPPROTO_UDP)) >= 0)
+    {
+      /* Disallow v4-in-v6 to allow host-based access checks */
+      if (setsockopt (sock, IPPROTO_IPV6, IPV6_V6ONLY,
+		  &i, sizeof(i)) == -1)
+	{
+	  log_msg ("ERROR: cannot disable v4-in-v6 on %s6 socket",
+	       nconf->nc_proto);
+	  return 1;
+	}
+      memset (&sin6, 0, sizeof (sin6));
+      sin6.sin6_family = AF_INET6;
+      if (my_port > 0)
+	sin6.sin6_port = htons (my_port);
+      sa = (struct sockaddr *)(void *)&sin6;
+
+      if (bindresvport_sa (sock, sa) == -1)
+	{
+	  if (my_port > 0)
+	    log_msg ("Cannot bind to reserved port %d (%s)",
+		     my_port, strerror (errno));
+	  else
+	    log_msg ("bindresvport failed: %s",
+		     strerror (errno));
+	  return 1;
+	}
+
+      if ((CallbackXprt = svc_dg_create (sock, 0, 0)) == NULL)
+	{
+	  log_msg ("terminating: cannot create rpcbind handle");
+	  return 1;
+	}
+
+      nconf = getnetconfigent ("udp6");
+      if (nconf == NULL)
+	{
+	  log_msg ("YPPUSH: getnetconfigent (\"udp6\") failed.");
+	  exit (1);
+	}
+      if (!svc_reg (CallbackXprt, CallbackProg, 1,
+		   yppush_xfrrespprog_1, nconf))
+	log_msg ("YPPUSH: couldn't register IPv6");
+      freenetconfigent (nconf);
+    }
+  else if (errno != EAFNOSUPPORT)
+    {
+      log_msg ("Cannot create UDP socket for AF_INET6: %s",
+	       strerror (errno));
+      return 1;
+    }
 
   switch (transid = fork ())
     {
@@ -491,16 +577,16 @@ yppush_foreach (const char *host)
       yppush_svc_run (server);
       exit (0);
     default:
-      close (CallbackXprt->xp_sock);
-      req.map_parms.domain = (char *) DomainName;
-      req.map_parms.map = (char *) current_map;
+      newreq.map_parms.domain = (char *) DomainName;
+      newreq.map_parms.map = (char *) current_map;
       /* local_hostname is correct since we have compared it
 	 with YP_MASTER_NAME.  */
-      req.map_parms.peer = local_hostname;
-      req.map_parms.ordernum = MapOrderNum;
-      req.transid = transid;
-      req.prog = CallbackProg;
-      req.port = CallbackXprt->xp_port;
+      newreq.map_parms.owner = local_hostname;
+      newreq.map_parms.ordernum = MapOrderNum;
+      newreq.transid = transid;
+      newreq.proto = CallbackProg;
+      // req.port = CallbackXprt->xp_port;
+      newreq.name = server;
 
       if (verbose_flag)
 	{
@@ -508,18 +594,37 @@ yppush_foreach (const char *host)
 	  if (verbose_flag > 1)
 	    {
 	      log_msg ("\t->target: %s", server);
-	      log_msg ("\t->domain: %s", req.map_parms.domain);
-	      log_msg ("\t->map: %s", req.map_parms.map);
-	      log_msg ("\t->tarnsid: %d", req.transid);
-	      log_msg ("\t->prog: %d", req.prog);
-	      log_msg ("\t->master: %s", req.map_parms.peer);
-	      log_msg ("\t->ordernum: %d", req.map_parms.ordernum);
+	      log_msg ("\t->domain: %s", newreq.map_parms.domain);
+	      log_msg ("\t->map: %s", newreq.map_parms.map);
+	      log_msg ("\t->tarnsid: %d", newreq.transid);
+	      log_msg ("\t->proto: %d", newreq.proto);
+	      log_msg ("\t->master: %s", newreq.map_parms.owner);
+	      log_msg ("\t->ordernum: %d", newreq.map_parms.ordernum);
+	      log_msg ("\t->name: %s", newreq.name);
 	    }
 	}
 
-      if (clnt_call (PushClient, YPPROC_XFR, (xdrproc_t) xdr_ypreq_xfr,
-		     (caddr_t) &req, (xdrproc_t) xdr_void, NULL,
-		     tv) != RPC_SUCCESS)
+
+      res = clnt_call (PushClient, YPPROC_NEWXFR, (xdrproc_t) xdr_ypreq_newxfr,
+		       (caddr_t) &newreq, (xdrproc_t) xdr_void, NULL, tv);
+
+      if (res == RPC_PROCUNAVAIL)
+	{
+	  struct ypreq_xfr oldreq;
+
+	  oldreq.map_parms.domain = (char *) DomainName;
+	  oldreq.map_parms.map = (char *) current_map;
+	  oldreq.map_parms.owner = local_hostname;
+	  oldreq.map_parms.ordernum = MapOrderNum;
+	  oldreq.transid = transid;
+	  oldreq.proto = CallbackProg;
+	  oldreq.port = 0; /* we don't really need that */
+
+	  res = clnt_call (PushClient, YPPROC_XFR, (xdrproc_t) xdr_ypreq_xfr,
+			   (caddr_t) &oldreq, (xdrproc_t) xdr_void, NULL, tv);
+	}
+
+      if (res != RPC_SUCCESS)
 	{
 	  log_msg ("YPPUSH: Cannot call YPPROC_XFR on host \"%s\"%s", server,
 		   clnt_sperror (PushClient, ""));
@@ -527,7 +632,7 @@ yppush_foreach (const char *host)
 	}
 
       waitpid (transid, &sock, 0);
-      svc_unregister (CallbackProg, 1);
+      svc_unreg (CallbackProg, 1);
       CallbackProg = 0;
       if (PushClient != NULL)
 	{
@@ -539,17 +644,74 @@ yppush_foreach (const char *host)
   return 0;
 }
 
+static char *
+get_canonical_hostname (const char *hostname)
+{
+#if USE_FQDN
+  struct addrinfo hints, *res0, *res1;
+  int error;
+  char *host = NULL;
+
+  memset (&hints, 0, sizeof(struct addrinfo));
+  hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
+#if 0
+  hints.ai_flags = AI_CANONNAME;  /* get the official name of the host */
+#endif
+
+  if ((error = getaddrinfo (hostname, NULL, &hints, &res0)))
+    {
+#if 0
+      printf ("getaddrinfo: %s\n", gai_strerror (error));
+#endif
+      return strdup (hostname);
+    }
+
+  res1 = res0;
+
+  while (res1)
+    {
+      char hostbuf[NI_MAXHOST];
+
+      if ((error = getnameinfo (res1->ai_addr, res1->ai_addrlen,
+                                (char *)&hostbuf, sizeof (hostbuf),
+                                NULL, 0, NI_NAMEREQD)) == 0)
+        {
+          host = strdup (hostbuf);
+          break;
+        }
+#if 0
+      else
+        printf ("getnameinfo: %s\n", gai_strerror (error));
+#endif
+
+      res1 = res1->ai_next;
+    }
+
+  if (host == NULL)
+    host = strdup (res0->ai_canonname);
+
+  freeaddrinfo (res0);
+
+  return host;
+#else
+  return strdup (hostname);
+#endif
+}
+
 static void
 sig_child (int sig UNUSED)
 {
   int status;
+  int save_errno = errno;
 
   while (waitpid (-1, &status, WNOHANG) > 0)
     {
       if (verbose_flag > 1)
-	log_msg ("Child %d exists", WEXITSTATUS (status));
+	log_msg ("Child %d exits", WEXITSTATUS (status));
       children--;
     }
+
+  errno = save_errno;
 }
 
 static inline void
@@ -643,6 +805,12 @@ main (int argc, char **argv)
 	      log_msg ("yppush cannot run in parallel with a fixed port");
 	      return 1;
 	    }
+	  if (my_port <= 0 || my_port > 0xffff) {
+	    /* Invalid port number */
+	    fprintf (stdout, "Warning: yppush: Invalid port %d (0x%x)\n",
+			my_port, my_port);
+	    my_port = -1;
+	  }
 	  break;
 	default:
 	  Usage (1);
@@ -675,23 +843,14 @@ main (int argc, char **argv)
       log_msg ("YPPUSH: Cannot determine local hostname");
       return 1;
     }
-#if USE_FQDN
   else
     {
-      struct hostent *hp;
+      char *cp;
 
-      if (!(hp = gethostbyname (local_hostname)))
-	{
-	  perror ("YPPUSH: gethostbyname()");
-	  log_msg ("YPPUSH: using not FQDN name");
-	}
-      else
-	{
-	  strncpy (local_hostname, hp->h_name, MAXHOSTNAMELEN);
-	  local_hostname[MAXHOSTNAMELEN] = '\0';
-	}
+      cp = get_canonical_hostname (local_hostname);
+      strncpy (local_hostname, cp, sizeof (local_hostname) -1);
     }
-#endif
+
 
   if (hostliste == NULL)
     {
@@ -699,11 +858,7 @@ main (int argc, char **argv)
 
       memset (&f, 0, sizeof f);
       f.foreach = add_slave_server;
-#ifdef OSF_KLUDGE
-      y = yp_all (DomainName, "ypservers", f);
-#else
       y = yp_all (DomainName, "ypservers", &f);
-#endif
       if (y && y != YP_NOMORE)
 	{
 	  log_msg ("Could not read ypservers map: %d %s", y, yperr_string (y));
